package net.froemling.bsremote;

import net.froemling.bsremote.GLRenderer;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import android.opengl.GLSurfaceView;
import android.os.Build;
import android.os.Bundle;
import android.os.SystemClock;
import android.provider.Settings.Secure;
import android.app.Activity;
import android.app.Dialog;
import android.content.Context;
import android.content.SharedPreferences;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.widget.RadioButton;
import android.widget.RelativeLayout;
import android.widget.SeekBar;
import android.widget.SeekBar.OnSeekBarChangeListener;
import android.widget.TextView;
import android.widget.Toast;
import android.util.Log;
import java.lang.Math;

class MyGLSurfaceView extends GLSurfaceView {
	public final static boolean debug = false;
	
	public final static String TAG = "SURF";

	private final GLRenderer _gl;
	private final GamePadActivity _gamePadActivity;

	int _dPadTouch = -1;
	
	int _prefsTouch = -1;
	int _quitTouch = -1;
	int _startTouch = -1;
	boolean _prefsHover;
	boolean _quitHover;
	boolean _startHover;
	
	boolean _dPadTouchIsMove;

	boolean _inited = false;
	float _buttonCenterX;
	float _buttonCenterY;
	float _dPadCenterX;
	float _dPadCenterY;

	float _dPadScale = 1.0f;
	float _buttonScale = 1.0f;
	float _dPadOffsX = 0.0f;
	float _dPadOffsY = 0.0f;
	float _buttonOffsX = 0.0f;
	float _buttonOffsY = 0.0f;
	float _dPadTouchStartX;
	float _dPadTouchStartY;
	
	float _sizeMin = 0.5f;
	float _sizeMax = 1.5f;

	float _dPadOffsXMin = -0.25f;
	float _dPadOffsXMax = 1.0f;
	float _dPadOffsYMin = -1.0f;
	float _dPadOffsYMax = 0.8f;

	float _buttonOffsXMin = -0.8f;
	float _buttonOffsXMax = 0.25f;
	float _buttonOffsYMin = -1.0f;
	float _buttonOffsYMax = 0.8f;

	String _dPadType = "floating";
	
	public MyGLSurfaceView(Context context) {
		super(context);

		//setSystemUiVisibility(View.SYSTEM_UI_FLAG_LOW_PROFILE);
		
		_gamePadActivity = (GamePadActivity) context;

		// Create an OpenGL ES 2.0 context.
		setEGLContextClientVersion(2);

		// Set the Renderer for drawing on the GLSurfaceView
		_gl = new GLRenderer(_gamePadActivity.getApplicationContext());

		// this seems to be necessary in the emulator..
		if (Build.FINGERPRINT.startsWith("generic"))
			setEGLConfigChooser(8, 8, 8, 8, 16, 0);
		setRenderer(_gl);

		// Render the view only when there is a change in the drawing data
		setRenderMode(GLSurfaceView.RENDERMODE_WHEN_DIRTY);

		SharedPreferences preferences = _gamePadActivity.getSharedPreferences("BSRemotePrefs", Context.MODE_PRIVATE);
		_dPadScale = preferences.getFloat("scale", 1.0f);
		_buttonScale = preferences.getFloat("buttonScale", 1.0f);
		_dPadScale = preferences.getFloat("dPadScale", 1.0f);
		_dPadOffsX = preferences.getFloat("dPadOffsX", 0.4f);
		_dPadOffsY = preferences.getFloat("dPadOffsY", 0.0f);
		_buttonOffsX = preferences.getFloat("buttonOffsX", -0.2f);
		_buttonOffsY = preferences.getFloat("buttonOffsY", 0.0f);
		_dPadType = preferences.getString("dPadType","floating");
		if (!(_dPadType.equals("floating") || _dPadType.equals("fixed"))) _dPadType = "floating";
	}

	private void _savePrefs() {
		// save this to prefs
		SharedPreferences preferences = _gamePadActivity.getSharedPreferences(
				"BSRemotePrefs", Context.MODE_PRIVATE);
		SharedPreferences.Editor editor = preferences.edit();
		editor.putFloat("buttonScale", _buttonScale);
		editor.putFloat("dPadScale", _dPadScale);
		editor.putFloat("dPadOffsX", _dPadOffsX);
		editor.putFloat("dPadOffsY", _dPadOffsY);
		editor.putFloat("buttonOffsX", _buttonOffsX);
		editor.putFloat("buttonOffsY", _buttonOffsY);
		editor.putString("dPadType", _dPadType);
		editor.commit();
	}

	private void _updateSizes() {

		// update button positions and whatnot
		float ratio = (float) getWidth() / getHeight();
		float height = 1.0f / ratio;

		float bWidth = 0.08f * _buttonScale;
		float bHeight = 0.08f * _buttonScale;
		_buttonCenterX = 0.95f - 0.1f * _buttonScale + _buttonOffsX * 0.2f;
		_buttonCenterY = height * 0.6f - 0.0f * _buttonScale - _buttonOffsY
				* 0.3f;
		_dPadCenterX = 0.0f + 0.1f * _dPadScale + _dPadOffsX * 0.2f;
		_dPadCenterY = height * 0.6f - 0.0f * _dPadScale - _dPadOffsY * 0.3f;
		float bSep = 0.1f * _buttonScale;

		_gl.quitButtonX = 0.06f * _buttonScale;
		_gl.quitButtonY = 0.035f * _buttonScale;
		_gl.quitButtonWidth = 0.1f * _buttonScale;
		_gl.quitButtonHeight = 0.05f * _buttonScale;

		_gl.prefsButtonX = 0.17f * _buttonScale;
		_gl.prefsButtonY = 0.035f * _buttonScale;
		_gl.prefsButtonWidth = 0.1f * _buttonScale;
		_gl.prefsButtonHeight = 0.05f * _buttonScale;
		
		_gl.startButtonX = 0.28f * _buttonScale;
		_gl.startButtonY = 0.035f * _buttonScale;
		_gl.startButtonWidth = 0.1f * _buttonScale;
		_gl.startButtonHeight = 0.05f * _buttonScale;

		_gl.throwButtonX = _buttonCenterX;
		_gl.throwButtonY = _buttonCenterY - bSep;
		_gl.throwButtonWidth = bWidth;
		_gl.throwButtonHeight = bHeight;

		_gl.punchButtonX = _buttonCenterX - bSep;
		_gl.punchButtonY = _buttonCenterY;
		_gl.punchButtonWidth = bWidth;
		_gl.punchButtonHeight = bHeight;

		_gl.bombButtonX = _buttonCenterX + bSep;
		_gl.bombButtonY = _buttonCenterY;
		_gl.bombButtonWidth = bWidth;
		_gl.bombButtonHeight = bHeight;

		_gl.jumpButtonX = _buttonCenterX;
		_gl.jumpButtonY = _buttonCenterY + bSep;
		_gl.jumpButtonWidth = bWidth;
		_gl.jumpButtonHeight = bHeight;

		_gl.joystickCenterX = _dPadCenterX;
		_gl.joystickCenterY = _dPadCenterY;
		_gl.joystickWidth = 0.2f * _dPadScale;
		_gl.joystickHeight = 0.2f * _dPadScale;

		_gl.joystickX = _dPadCenterX;
		_gl.joystickY = _dPadCenterY;

		if (!_inited) {
			_inited = true;
		}
	}

	public void onSizeChanged(int w, int h, int oldw, int oldh) {
		_updateSizes();
	}

	private boolean _pointInBox(float x, float y, float bx, float by, float sx,
			float sy) {
		if (x >= bx - 0.5f * sx && x <= bx + 0.5 * sx && y >= by - 0.5 * sy
				&& y <= by + 0.5 * sy)
			return true;
		else
			return false;
	}

	void _updateButtonsForTouches(MotionEvent event) {
		boolean punchHeld = false;
		boolean throwHeld = false;
		boolean jumpHeld = false;
		boolean bombHeld = false;
		float mult = 1.0f / getWidth();
		final int pointerCount = event.getPointerCount();
		for (int i = 0; i < pointerCount; i++) {

			// ignore touch-up events
			int actionPointerIndex = event.getActionIndex();
			int action = event.getActionMasked();
			// int fingerID = event.getPointerId(actionPointerIndex);
			if ((action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_POINTER_UP)
					&& i == actionPointerIndex)
				continue;

			int touch = event.getPointerId(i);

			// ignore dpad touch
			if (touch == _dPadTouch) {
				continue;
			}

			float s = 4.0f / _buttonScale;
			// get the point in button-center-coords
			float x = event.getX(i) * mult;
			float y = event.getY(i) * mult;
			float bx = (x - _buttonCenterX) * s;
			float by = (y - _buttonCenterY) * s;

			float threshold = 0.3f;
			float pbx, pby;
			float len;
			float punchLen, jumpLen, throwLen, bombLen;

			// punch
			pbx = (x - _gl.punchButtonX) * s;
			pby = (y - _gl.punchButtonY) * s;
			punchLen = len = (float) Math.sqrt(pbx * pbx + pby * pby);
			if (len < threshold) {
				punchHeld = true;
			}

			// throw
			pbx = (x - _gl.throwButtonX) * s;
			pby = (y - _gl.throwButtonY) * s;
			throwLen = len = (float) Math.sqrt(pbx * pbx + pby * pby);
			if (len < threshold) {
				throwHeld = true;
			}

			// jump
			pbx = (x - _gl.jumpButtonX) * s;
			pby = (y - _gl.jumpButtonY) * s;
			jumpLen = len = (float) Math.sqrt(pbx * pbx + pby * pby);
			if (len < threshold) {
				jumpHeld = true;
			}

			// bomb
			pbx = (x - _gl.bombButtonX) * s;
			pby = (y - _gl.bombButtonY) * s;
			bombLen = len = (float) Math.sqrt(pbx * pbx + pby * pby);
			if (len < threshold) {
				bombHeld = true;
			}

			// how much larger than the button/dpad areas we should count touch
			// events in
			float buttonBuffer = 2.0f;
			// ok now lets take care of fringe areas and non-moved touches
			// a touch in our button area should *always* affect at least one
			// button
			// ..so lets find the closest button && press it.
			// this will probably coincide with what we just set above but thats
			// ok.
			if (x > 0.5 && bx > -1.0 * buttonBuffer && bx < 1.0 * buttonBuffer
					&& by > -1.0 * buttonBuffer && by < 1.0 * buttonBuffer) {
				if (punchLen < throwLen && punchLen < jumpLen
						&& punchLen < bombLen)
					punchHeld = true;
				else if (throwLen < punchLen && throwLen < jumpLen
						&& throwLen < bombLen)
					throwHeld = true;
				else if (jumpLen < punchLen && jumpLen < throwLen
						&& jumpLen < bombLen)
					jumpHeld = true;
				else
					bombHeld = true;
			}
		}

		boolean throwWasHeld = ((_gamePadActivity._buttonState & GamePadActivity.BS_REMOTE_STATE_THROW) != 0);
		boolean jumpWasHeld = ((_gamePadActivity._buttonState & GamePadActivity.BS_REMOTE_STATE_JUMP) != 0);
		boolean punchWasHeld = ((_gamePadActivity._buttonState & GamePadActivity.BS_REMOTE_STATE_PUNCH) != 0);
		boolean bombWasHeld = ((_gamePadActivity._buttonState & GamePadActivity.BS_REMOTE_STATE_BOMB) != 0);

		// send press events for non-held ones we're now over
		if (!throwWasHeld && throwHeld) {
			_handleThrowPress();
		}
		if (throwWasHeld && !throwHeld) {
			_handleThrowRelease();
		}

		// send press events for non-held ones we're now over
		if ((!punchWasHeld) && punchHeld) {
			_handlePunchPress();
		}
		if (punchWasHeld && !punchHeld) {
			_handlePunchRelease();
		}

		// send press events for non-held ones we're now over
		if ((!bombWasHeld) && bombHeld) {
			_handleBombPress();
		}
		if (bombWasHeld && !bombHeld) {
			_handleBombRelease();
		}

		// send press events for non-held ones we're now over
		if ((!jumpWasHeld) && jumpHeld) {
			_handleJumpPress();
		}
		if (jumpWasHeld && !jumpHeld) {
			_handleJumpRelease();
		}

	}

	private void _handleThrowPress() {
		_gamePadActivity._buttonState |= GamePadActivity.BS_REMOTE_STATE_THROW;
		_gamePadActivity._doStateChange(false);
		_gl.throwPressed = true;
	}

	private void _handleThrowRelease() {
		_gamePadActivity._buttonState &= ~GamePadActivity.BS_REMOTE_STATE_THROW;
		_gamePadActivity._doStateChange(false);
		_gl.throwPressed = false;
	}

	private void _handleBombPress() {
		_gamePadActivity._buttonState |= GamePadActivity.BS_REMOTE_STATE_BOMB;
		_gamePadActivity._doStateChange(false);
		_gl.bombPressed = true;
	}

	private void _handleBombRelease() {
		_gamePadActivity._buttonState &= ~GamePadActivity.BS_REMOTE_STATE_BOMB;
		_gamePadActivity._doStateChange(false);
		_gl.bombPressed = false;
	}

	private void _handleJumpPress() {
		_gamePadActivity._buttonState |= GamePadActivity.BS_REMOTE_STATE_JUMP;
		_gamePadActivity._doStateChange(false);
		_gl.jumpPressed = true;
	}

	private void _handleJumpRelease() {
		_gamePadActivity._buttonState &= ~GamePadActivity.BS_REMOTE_STATE_JUMP;
		_gamePadActivity._doStateChange(false);
		_gl.jumpPressed = false;
	}

	private void _handlePunchPress() {
		_gamePadActivity._buttonState |= GamePadActivity.BS_REMOTE_STATE_PUNCH;
		_gamePadActivity._doStateChange(false);
		_gl.punchPressed = true;
	}

	private void _handlePunchRelease() {
		_gamePadActivity._buttonState &= ~GamePadActivity.BS_REMOTE_STATE_PUNCH;
		_gamePadActivity._doStateChange(false);
		_gl.punchPressed = false;
	}

	public void doPrefs(){

		
		Dialog d = new Dialog(_gamePadActivity);
		d.setContentView(R.layout.prefs);
		
		SeekBar seekbar;
		seekbar = (SeekBar) d.findViewById(R.id.seekBarButtonSize);
		seekbar.setProgress((int) (100.0f * (_buttonScale - _sizeMin) / (_sizeMax - _sizeMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_buttonScale = _sizeMin + (_sizeMax - _sizeMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		seekbar = (SeekBar) d.findViewById(R.id.seekBarDPadSize);
		seekbar.setProgress((int) (100.0f * (_dPadScale - _sizeMin) / (_sizeMax - _sizeMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_dPadScale = _sizeMin + (_sizeMax - _sizeMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		seekbar = (SeekBar) d
				.findViewById(R.id.seekBarButtonPosition1);
		seekbar.setProgress((int) (100.0f * (_buttonOffsX - _buttonOffsXMin) / (_buttonOffsXMax - _buttonOffsXMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_buttonOffsX = _buttonOffsXMin
						+ (_buttonOffsXMax - _buttonOffsXMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		seekbar = (SeekBar) d
				.findViewById(R.id.seekBarButtonPosition2);
		seekbar.setProgress((int) (100.0f * (_buttonOffsY - _buttonOffsYMin) / (_buttonOffsYMax - _buttonOffsYMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_buttonOffsY = _buttonOffsYMin
						+ (_buttonOffsYMax - _buttonOffsYMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		seekbar = (SeekBar) d
				.findViewById(R.id.seekBarDPadPosition1);
		seekbar.setProgress((int) (100.0f * (_dPadOffsX - _dPadOffsXMin) / (_dPadOffsXMax - _dPadOffsXMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_dPadOffsX = _dPadOffsXMin
						+ (_dPadOffsXMax - _dPadOffsXMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		seekbar = (SeekBar) d
				.findViewById(R.id.seekBarDPadPosition2);
		seekbar.setProgress((int) (100.0f * (_dPadOffsY - _dPadOffsYMin) / (_dPadOffsYMax - _dPadOffsYMin)));
		seekbar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
			public void onProgressChanged(SeekBar seekBar,
					int progress, boolean fromUser) {
				_dPadOffsY = _dPadOffsYMin
						+ (_dPadOffsYMax - _dPadOffsYMin)
						* (progress / 100.0f);
				_updateSizes();
				requestRender();
			}

			public void onStartTrackingTouch(SeekBar seekBar) {
			}

			public void onStopTrackingTouch(SeekBar seekBar) {
				_savePrefs();
			}
		});
		
		RadioButton dPadFloatingButton = (RadioButton)d.findViewById(R.id.radioButtonDPadFloating);
		RadioButton dPadFixedButton = (RadioButton)d.findViewById(R.id.radioButtonDPadFixed);

		if (_dPadType.equals("floating")) dPadFloatingButton.setChecked(true);
		else dPadFixedButton.setChecked(true);
		
		dPadFloatingButton.setOnClickListener(new RadioButton.OnClickListener(){
			@Override
			public void onClick(View v) {
				_dPadType = "floating";
				_savePrefs();
			}});
		dPadFixedButton.setOnClickListener(new RadioButton.OnClickListener(){
			@Override
			public void onClick(View v) {
				_dPadType = "fixed";
				_savePrefs();
			}});

		d.setTitle("Preferences");
		d.show();

	
	}
	// Touch events
	public boolean onTouchEvent(MotionEvent event) {
		final int pointerCount = event.getPointerCount();
		int actionPointerIndex = event.getActionIndex();
		int action = event.getActionMasked();

		float mult = 1.0f / getWidth();

		// touch margin
		float tm = 1.4f;
		float tm2 = 2.4f;

		for (int i = 0; i < pointerCount; i++) {
			float x = event.getX(i) * mult;
			float y = event.getY(i) * mult;
			int fingerID = event.getPointerId(i);

			if ((action == MotionEvent.ACTION_DOWN || action == MotionEvent.ACTION_POINTER_DOWN)
					&& (actionPointerIndex == i)) {

				// prefs presses
				if (_pointInBox(x, y, _gl.prefsButtonX, _gl.prefsButtonY,
						_gl.prefsButtonWidth * tm, _gl.prefsButtonHeight * tm2)){
					_prefsTouch = fingerID;
					_prefsHover = true;
					_gl.prefsButtonPressed = true;
				}

				// check for a quit press
				else if (_pointInBox(x, y, _gl.quitButtonX, _gl.quitButtonY,
						_gl.quitButtonWidth * tm, _gl.quitButtonHeight * tm2)) {
					_quitTouch = fingerID;
					_quitHover = true;
					_gl.quitButtonPressed = true;
				}
				// check for a start press
				else if (_pointInBox(x, y, _gl.startButtonX, _gl.startButtonY,
						_gl.startButtonWidth * tm, _gl.startButtonHeight * tm2)) {
					_startTouch = fingerID;
					_startHover = true;
					_gl.startButtonPressed = true;
				}
				// check for a dpad touch
				else if (x < 0.5) {
					_dPadTouch = fingerID;
					_dPadTouchIsMove = false;
					// in fixed joystick mode we want touches to count towards
					// joystick motion immediately; not just after they move
					if (_dPadType.equals("fixed"))
						_dPadTouchIsMove = true;
					
					_dPadTouchStartX = x;
					_dPadTouchStartY = y;
					_gl.joystickX = x;//_gl.joystickCenterX = x;
					_gl.joystickY = y;//_gl.joystickCenterY = y;
					_gl.thumbPressed = true;

				}
			}
			// handle existing button touches
			if (fingerID == _quitTouch){
				// update position
				_quitHover = _gl.quitButtonPressed = (_pointInBox(x, y, _gl.quitButtonX, _gl.quitButtonY,
						_gl.quitButtonWidth * tm, _gl.quitButtonHeight * tm2));
				// if the touch is ending...
				if ((action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_POINTER_UP)
						&& (actionPointerIndex == i)) {
					_quitTouch = -1;
					//_quitHover = false;
					_gl.quitButtonPressed = false;
					if (_quitHover){
						_gamePadActivity.finish();
					}
				}
			}
			else if (fingerID == _prefsTouch){
				// update position
				_prefsHover = _gl.prefsButtonPressed = (_pointInBox(x, y, _gl.prefsButtonX, _gl.prefsButtonY,
						_gl.prefsButtonWidth * tm, _gl.prefsButtonHeight * tm2));
				// if the touch is ending...
				if ((action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_POINTER_UP)
						&& (actionPointerIndex == i)) {
					_prefsTouch = -1;
					//_prefsHover = false;
					_gl.prefsButtonPressed = false;
					if (_prefsHover){
						doPrefs();
					}
				}
			}
			else if (fingerID == _startTouch){
				// update position
				_startHover = _gl.startButtonPressed = (_pointInBox(x, y, _gl.startButtonX, _gl.startButtonY,
						_gl.startButtonWidth * tm, _gl.startButtonHeight * tm2));
				// if the touch is ending...
				if ((action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_POINTER_UP)
						&& (actionPointerIndex == i)) {
					_startTouch = -1;
					//_startHover = false;
					_gl.startButtonPressed = false;
					if (_startHover){
						// send 2 state-changes (start-down and start-up)
						_gamePadActivity._buttonState |= GamePadActivity.BS_REMOTE_STATE_MENU;
						_gamePadActivity._doStateChange(false);
						_gamePadActivity._buttonState &= ~GamePadActivity.BS_REMOTE_STATE_MENU;
						_gamePadActivity._doStateChange(false);
					}
				}
			}


			// if its our existing dpad-touch
			if (fingerID == _dPadTouch) {

				// if we've moved away from the initial touch position we no longer
				// consider it for a direction tap
				if (!_dPadTouchIsMove && (Math.abs(x - _dPadTouchStartX) > 0.01 || Math.abs(y-_dPadTouchStartY) > 0.01)){
					_dPadTouchIsMove = true;
					_gl.joystickCenterX = _dPadTouchStartX;
					_gl.joystickCenterY = _dPadTouchStartY;
				}
				// if its moved, pass it along as a joystick event
				if (_dPadTouchIsMove){
					float s = 30.0f / _dPadScale;
					float xVal = (x - _gl.joystickCenterX) * s;
					float yVal = (y - _gl.joystickCenterY) * s;
					float xValClamped = xVal;
					float yValClamped = yVal;

				    // keep our H/V values within a unit box
				    // (originally I clamped length to 1 but as a result our diagonal
				    // running speed was less than other analog controllers..)

				    if (xValClamped > 1.0f){
				        float m = 1.0f/xValClamped;
				        xValClamped *= m;
				        yValClamped *= m;
				    }
				    else if (xValClamped < -1.0f){
				        float m = -1.0f/xValClamped;
				        xValClamped *= m;
				        yValClamped *= m;
				    }
				    if (yValClamped > 1.0f){
				        float m = 1.0f/yValClamped;
				        xValClamped *= m;
				        yValClamped *= m;
				    }
				    else if (yValClamped < -1.0f){
				        float m = -1.0f/yValClamped;
				        xValClamped *= m;
				        yValClamped *= m;
				    }

//					if (xValClamped < -1.0f)
//						xValClamped = -1.0f;
//					else if (xValClamped > 1.0f)
//						xValClamped = 1.0f;
//					if (yValClamped < -1.0f)
//						yValClamped = -1.0f;
//					else if (yValClamped > 1.0f)
//						yValClamped = 1.0f;

					_gl.joystickX = _gl.joystickCenterX + xVal / s;
					_gl.joystickY = _gl.joystickCenterY + yVal / s;

					// if its moved far enough away from center, have the dpad
					// follow it (in floating mode)
					// in fixed mode just clamp distance
					float dist = (float) Math.sqrt(xVal * xVal + yVal * yVal);
					if (_dPadType.equals("floating")){
						if (dist > 1.5f) {
							float sc = 1.5f / dist;
							_gl.joystickCenterX = _gl.joystickX - sc * (xVal / s);
							_gl.joystickCenterY = _gl.joystickY - sc * (yVal / s);
						}
					}
					else if (_dPadType.equals("fixed")){
						if (dist > 1.01f) {
							float sc = 1.01f / dist;
							_gl.joystickX = _gl.joystickCenterX + sc * (xVal / s);
							_gl.joystickY = _gl.joystickCenterY + sc * (yVal / s);
						}

					}

					_gamePadActivity._dPadStateH = xValClamped;
					_gamePadActivity._dPadStateV = yValClamped;
					_gamePadActivity._doStateChange(false);
				}

				// if the touch is ending...
				if ((action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_POINTER_UP)
						&& (actionPointerIndex == i)) {
					
					// if we hadnt moved yet, lets pass it along as a quick tap/release
					// (useful for navigating menus)
					if (!_dPadTouchIsMove){
						float toRight = x-_gl.joystickCenterX;
						float toLeft = _gl.joystickCenterX-x;
						float toBottom = y-_gl.joystickCenterY;
						float toTop = _gl.joystickCenterY-y;
						// right
						if (toRight > toLeft && toRight > toTop && toRight > toBottom){
							_gamePadActivity._dPadStateH = 1.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);
						}
						// left
						else if (toLeft > toRight && toLeft > toTop && toLeft > toBottom){
							_gamePadActivity._dPadStateH = -1.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);
						}
						else if (toTop > toRight && toTop > toLeft && toTop > toBottom){
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = -1.0f;
							_gamePadActivity._doStateChange(false);
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);

						}
						else{
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = 1.0f;
							_gamePadActivity._doStateChange(false);
							_gamePadActivity._dPadStateH = 0.0f;
							_gamePadActivity._dPadStateV = 0.0f;
							_gamePadActivity._doStateChange(false);
						}
					}
					_dPadTouch = -1;
					_gl.thumbPressed = false;
					_gl.joystickCenterX = _dPadCenterX;
					_gl.joystickCenterY = _dPadCenterY;
					_gl.joystickX = _gl.joystickCenterX;
					_gl.joystickY = _gl.joystickCenterY;
					_gamePadActivity._dPadStateH = 0;
					_gamePadActivity._dPadStateV = 0;
					_gamePadActivity._doStateChange(false);

				}
			}
		}
		_updateButtonsForTouches(event);
		requestRender();
		return true;
	}
}

public class GamePadActivity extends Activity {

	public final static boolean debug = false;
	
	public final static String TAG = "GPA";
	private WorkerThread _readThread;
	private WorkerThread _processThread;
	private Timer _processTimer;
	private Timer _shutDownTimer;
	private DatagramSocket _socket;
	private InetAddress[] _addrs;
	private int _port;
	private int _requestID;
	private byte _id;
	long _shutDownStartTime;
	private boolean _dead = false;
	private long _lastLagUpdateTime;
	float _averageLag;
	TextView _lagMeter;

	int _uniqueID;
	
	static final int BS_REMOTE_ERROR_VERSION_MISMATCH = 0;
	static final int BS_REMOTE_ERROR_GAME_SHUTTING_DOWN = 1;
	static final int BS_REMOTE_ERROR_NOT_ACCEPTING_CONNECTIONS = 2;
	static final int BS_REMOTE_ERROR_NOT_CONNECTED = 3;

	static final int REMOTE_MSG_ID_REQUEST = 2;
	static final int REMOTE_MSG_ID_RESPONSE = 3;
	static final int REMOTE_MSG_DISCONNECT = 4;
	static final int REMOTE_MSG_STATE = 5;
	static final int REMOTE_MSG_STATE_ACK = 6;
	static final int REMOTE_MSG_DISCONNECT_ACK = 7;

	static final int BS_REMOTE_STATE_PUNCH = 1 << 0;
	static final int BS_REMOTE_STATE_JUMP = 1 << 1;
	static final int BS_REMOTE_STATE_THROW = 1 << 2;
	static final int BS_REMOTE_STATE_BOMB = 1 << 3;
	static final int BS_REMOTE_STATE_MENU = 1 << 4;

	// (bits 6-10 are d-pad h-value and bits 11-15 are dpad v-value)
	static final int BS_REMOTE_STATE_HOLD_POSITION = 1 << 15;
	private int _nextState;
	private boolean _connected = false;
	private boolean _shouldPrintConnected;
	private int _requestedState = 0;
	private boolean _shuttingDown;
	private long _lastNullStateTime = 0;
	private long[] _stateBirthTimes;
	private long[] _stateLastSentTimes;
	short _buttonState = 0;
	float _dPadStateH = 0.0f;
	float _dPadStateV = 0.0f;
	private long _lastSentState = 0;
	short[] _states;
	float _currentLag;

	// private GameView mSurface;
	private MyGLSurfaceView mGLView;
	private ViewGroup mLayout;

	private String[] _addrsRaw;

	private boolean _newStyle;

	@Override
	protected void onCreate(Bundle savedInstanceState) {

		// clients use a random request ID to differentiate themselves
		// (probably is a better way to do this..)
		long currentTime = SystemClock.uptimeMillis();
		_requestID = (int) currentTime % 10000;
		_id = -1; // aint got one yet
		_stateBirthTimes = new long[256];
		_stateLastSentTimes = new long[256];
		_states = new short[256];

		// if we reconnect we may get acks for states we didnt send..
		// so lets set everything to current time to avoid screwing up
		// our lag-meter
		long curTime = SystemClock.uptimeMillis();
		for (int i = 0; i < 256; i++){
			_stateBirthTimes[i] = curTime;
			_stateLastSentTimes[i] = 0;
		}
		
		Bundle extras = getIntent().getExtras();
		if (extras != null) {
			_newStyle = extras.getBoolean("newStyle");
			_port = extras.getInt("connectPort");
			try {
				_socket = new DatagramSocket();
			} catch (SocketException e) {
				LogThread.log("Error setting up gamepad socket",e);
			}
			_addrsRaw = extras.getStringArray("connectAddrs");
		}
		
		// read or create our random unique ID; we tack this onto our
		// android device identifier just in case its not actually unique
		// (as apparently was the case with nexus-2 or some other device)

		SharedPreferences preferences = getSharedPreferences("BSRemotePrefs", Context.MODE_PRIVATE);
		_uniqueID = preferences.getInt("uniqueId",0);
		if (_uniqueID == 0){
			while (_uniqueID == 0){
				_uniqueID = new Random().nextInt() & 0xFFFF;
			}
			SharedPreferences.Editor editor = preferences.edit();
			editor.putInt("uniqueId",_uniqueID);
			editor.commit();
			
		}

		_processThread = new WorkerThread();
		_processThread.start();

		_readThread = new WorkerThread();
		_readThread.start();

		// all the read-thread does is wait for data to come in
		// and pass it to the process-thread
		_readThread.doRunnable(new Runnable() {
			public void run() {
				while (true) {
					try {
						byte[] buf = new byte[10];
						DatagramPacket packet = new DatagramPacket(buf,
								buf.length);
						_socket.receive(packet);
						abstract class PacketRunnable implements Runnable {
							DatagramPacket p;

							PacketRunnable(DatagramPacket pIn) {
								p = pIn;
							}

							public abstract void run();
						}
						_processThread.doRunnable(new PacketRunnable(packet) {
							public void run() {
								GamePadActivity.this._readFromSocket(p);
							}
						});

					} catch (IOException e) {
						// assuming this means the socket is closed..
						if (debug) Log.v(TAG, "READ THREAD DYING");
						_readThread.getLooper().quit();
						_readThread = null;
						break;
					} catch (ArrayIndexOutOfBoundsException e) {
						LogThread.log("Got excessively sized datagram packet", e);
					}
				}
			}
		});

		super.onCreate(savedInstanceState);

		// Create a GLSurfaceView instance and set it
		// as the ContentView for this Activity
		mGLView = new MyGLSurfaceView(this);

		mLayout = new RelativeLayout(this);
		mLayout.addView(mGLView);

		_lagMeter = new TextView(this);
		_lagMeter.setTextColor(0xFF00FF00);
		_lagMeter.setText("Lag: --");
		RelativeLayout.LayoutParams params = new RelativeLayout.LayoutParams(
				RelativeLayout.LayoutParams.WRAP_CONTENT,
				RelativeLayout.LayoutParams.WRAP_CONTENT);
		params.addRule(RelativeLayout.ALIGN_PARENT_BOTTOM, RelativeLayout.TRUE);
		params.addRule(RelativeLayout.CENTER_IN_PARENT, RelativeLayout.TRUE);

		params.bottomMargin = 20;
		mLayout.addView(_lagMeter, params);

		setContentView(mLayout);
	}

	@Override
	public void onDestroy() {
		

		super.onDestroy();
		if (debug) Log.v(TAG, "onDestroy()");

		_shuttingDown = true;
		_shutDownStartTime = SystemClock.uptimeMillis();

		// create our shutdown timer.. this will keep us
		// trying to disconnect cleanly with the server until
		// we get confirmation or we give up
		// tell our worker thread to start its update timer
		_processThread.doRunnable(new Runnable() {
			public void run() {
				if (debug) Log.v(TAG, "CREATING SHUTDOWN TIMER...");
				assert (_shutDownTimer == null);
				_shutDownTimer = new Timer();
				_shutDownTimer.schedule(new TimerTask() {
					@Override
					public void run() {
						// when this timer fires, tell our process
						// thread to run
						_processThread.doRunnable(new Runnable() {
							public void run() {
								GamePadActivity.this._process();
							}
						});
					}
				}, 0, 100);
			}
		});
	}

	@Override
	protected void onStart() {
		
		super.onStart();
		if (debug) Log.v(TAG, "GPA onStart()");

		// tell our worker thread to start its update timer
		_processThread.doRunnable(new Runnable() {
			public void run() {
				// kick off an id request... (could just wait for the process
				// timer to do this)
				if (_id == -1)
					GamePadActivity.this._sendIdRequest();

				if (debug) Log.v(TAG, "CREATING PROCESS TIMER...");
				_processTimer = new Timer();
				_processTimer.schedule(new TimerTask() {
					@Override
					public void run() {
						// when this timer fires, tell our process
						// thread to run
						_processThread.doRunnable(new Runnable() {
							public void run() {
								GamePadActivity.this._process();
							}
						});
					}
				}, 0, 100);
			}
		});

	}

	protected void onStop() {
		super.onStop();

		_processThread.doRunnable(new Runnable() {
			public void run() {
				_processTimer.cancel();
				_processTimer.purge();
			}
		});

	}

	private void _process() {
		assert (Thread.currentThread() == _processThread);

		// if any of these happen after we're dead totally ignore them
		if (_dead)
			return;

		
		long t = SystemClock.uptimeMillis();

		// if we're shutting down but are still connected, keep sending
		// disconnects out
		if (_shuttingDown) {

			boolean finishUsOff = false;

			// shoot off another disconnect notice
			// once we're officially disconnected we can die
			if (!_connected) {
				finishUsOff = true;
			} else {
				if (_id != -1) {
					byte[] data = new byte[2];
					data[0] = REMOTE_MSG_DISCONNECT;
					data[1] = (byte) _id;
					try {
						_socket.send(new DatagramPacket(data, data.length));
					} catch (IOException e) {
						LogThread.log("", e);
					}
				}
				// just give up after a short while
				if (t - _shutDownStartTime > 5000) {
					finishUsOff = true;
				}
			}
			if (finishUsOff && !_dead) {
				_dead = true;
				// ok we're officially dead.. clear out our threads/timers/etc
				_shutDownTimer.cancel();
				_shutDownTimer.purge();
				_processThread.getLooper().quit();
				_processThread = null;
				// this should kill our read-thread
				_socket.close();
				return;
			}
		}

		// if we've got states we havn't heard an ack for yet, keep shipping 'em
		// out
		int stateDiff = (_requestedState - _nextState) & 0xFF;
		assert (stateDiff >= 0);

		// if they've requested a state we don't have yet, we don't need to
		// resend anything
		if (stateDiff < 128) {
			// .. however we wanna shoot states at the server every now and then
			// even if we have no new states,
			// to keep from timing out and such..
			if (t - _lastNullStateTime > 3000) {
				_doStateChange(true);
				_lastNullStateTime = t;
			}
		}
		// ok we've got at least one state we havn't heard confirmation for
		// yet.. lets ship 'em out..
		else {
			_shipUnAckedStates();
		}

		// if we don't have an ID yet, keep sending off those requests...
		if (_id == -1) {
			_sendIdRequest();
		}

		// update our lag meter every so often
		if (t - _lastLagUpdateTime > 2000) {
			float smoothing = 0.5f;
			_averageLag = smoothing * _averageLag + (1.0f - smoothing)
					* _currentLag;

			// lets show half of our the round-trip time as lag.. (the actual
			// delay
			// in-game is just our packets getting to them; not the round trip)
			runOnUiThread(new Runnable() {
				public void run() {
					if (_shouldPrintConnected) {
						_lagMeter.setTextColor(0xFF00FF00);
						_lagMeter.setText("Connected.");
						_shouldPrintConnected = false;

					} else if (_connected) {
						// convert from millisecs to seconds
						float val = (_averageLag * 0.5f) / 1000.0f;
						if (val < 0.1)
							_lagMeter.setTextColor(0xFF88FF88);
						else if (val < 0.2)
							_lagMeter.setTextColor(0xFFFFB366);
						else
							_lagMeter.setTextColor(0xFFFF6666);
						_lagMeter.setText(String.format("Lag: %.2f seconds",
								val));
					}
					// connecting...
					else {
						_lagMeter.setTextColor(0xFFFF8800);
						_lagMeter.setText("Connecting...");
					}
				}
			});
			_currentLag = 0.0f;
			_lastLagUpdateTime = t;
		}
	}

	private void _shipUnAckedStates() {
		// if we don't have an id yet this is moot..
		if (_id == -1)
			return;

		long curTime = SystemClock.uptimeMillis();

		// ok we need to ship out everything from their last requested state
		// to our current state.. (clamping at a reasonable value)
		if (_id != -1) {

			int statesToSend = (_nextState - _requestedState) & 0xFF;
			assert (statesToSend >= 0);
			if (statesToSend > 11)
				statesToSend = 11;
			if (statesToSend < 1)
				return;

			byte[] data = new byte[100];
			data[0] = REMOTE_MSG_STATE;
			data[1] = _id;
			data[2] = (byte) statesToSend; // number of states we have here

			int s = (_nextState - statesToSend) & 0xFF;
			assert (s >= 0);
			if (debug) Log.v(TAG, "SENDING " + statesToSend + " STATES FROM: " + s);
			data[3] = (byte) (s & 0xFF); // starting index

			// pack em in
			int index = 4;
			for (int i = 0; i < statesToSend; i++) {
				data[index++] = (byte) (_states[s] & 0xFF);
				data[index++] = (byte) (_states[s] >> 8);
				_stateLastSentTimes[s] = curTime;
				s = (s + 1) % 256;
			}
			if (_connected) {
				try {
					_socket.send(new DatagramPacket(data, 4 + 2 * statesToSend));
				} catch (IOException e) {
					
					//LogThread.log("Just testing a string",e);
					
					// if anything went wrong here, assume the game just shut down
					runOnUiThread(new Runnable() {
						public void run(){
							String msg = "The game has shut down.";
							Context context = getApplicationContext();
							int duration = Toast.LENGTH_LONG;
							Toast toast = Toast.makeText(context,msg,duration);
							toast.show();
						}

					});
					_connected = false;
					finish();

				}
			}
		}
	}

	void _doStateChange(boolean force) {

		//Log.v(TAG,"S "+_dPadStateH+" "+_dPadStateV);
		
		// compile our state value
		short s = _buttonState; // buttons
		s |= (_dPadStateH > 0 ? 1 : 0) << 5; // sign bit
		s |= ((int) (Math.round(Math.min(1.0, Math.abs(_dPadStateH)) * 15.0))) << 6; // mag
		s |= (_dPadStateV > 0 ? 1 : 0) << 10; // sign bit
		s |= ((int) (Math.round(Math.min(1.0, Math.abs(_dPadStateV)) * 15.0))) << 11; // mag

		// if our compiled state value hasn't changed, don't send.
		// (analog joystick noise can send a bunch of redundant states through
		// here)
		// The exception is if forced is true, which is the case with packets
		// that
		// double as keepalives.
		if ((s == _lastSentState) && (!force))
			return;

		_stateBirthTimes[_nextState] = SystemClock.uptimeMillis();
		_stateLastSentTimes[_nextState] = 0;

		if (debug) Log.v(TAG, "STORING NEXT STATE: " + _nextState);
		_states[_nextState] = s;
		_nextState = (_nextState + 1) % 256;
		_lastSentState = s;

		// if we're pretty up to date as far as state acks, lets go ahead
		// and send out this state immediately..
		// (keeps us nice and responsive on low latency networks)
		int unackedCount = (_nextState - _requestedState) & 0xFF; // upcast to
																	// get
																	// unsigned
		assert (unackedCount >= 0);
		if (unackedCount < 3) {
			_shipUnAckedStates();
		}
	}

	private void _sendIdRequest() {
		assert (!_connected);
		assert (_id == -1);
		assert (Thread.currentThread() == _processThread);

		// get our unique identifier to tack onto the end of our name
		// (so if we drop our connection and reconnect we can be reunited with our old
		// dude instead of leaving a zombie)
		String android_id = Secure.getString(getApplicationContext().getContentResolver(),Secure.ANDROID_ID);
		
		// on new-style connections we include unique id info in our name so we can be re-connected
		// if we disconnect
		String deviceName;
		if (_newStyle) deviceName = "Android remote (" + android.os.Build.MODEL + ")#"+android_id+_uniqueID;
		else deviceName = "Android remote (" + android.os.Build.MODEL + ")";
		
		int dLen = deviceName.length();
		if (dLen > 64)
			dLen = 64;

		// send a hello on all our addrs
		//Log.v(TAG, "Sending ID Request");
		byte[] data = new byte[128];
		data[0] = REMOTE_MSG_ID_REQUEST;
		data[1] = 121; // protocol version
		data[2] = (byte) (_requestID & 0xFF);
		data[3] = (byte) (_requestID >> 8);
		data[4] = 0; // unused..
		int len = 5;
		for (int i = 0; i < dLen; i++) {
			int c = deviceName.charAt(i);
			if (c < 256) {
				data[5 + i] = (byte) c;
				len++;
			}
		}
		// resolve our addresses if we havn't yet (couldnt do this in main thread)
		if (_addrs == null){
			int i = 0;
			_addrs = new InetAddress[_addrsRaw.length];
			for (String a : _addrsRaw) {
				try {
					_addrs[i] = InetAddress.getByName(a);
				} catch (UnknownHostException e) {
					LogThread.log("",e);
				}
				i++;
			}
		}
		for (InetAddress addr : _addrs) {
			DatagramPacket packet = new DatagramPacket(data, len, addr, _port);
			try {
				_socket.send(packet);
			} catch (IOException e) {
				LogThread.log("Error on ID-request send", e);
				Log.e(TAG, "Error on ID-request send: " + e.getMessage());
				e.printStackTrace();
			}
		}
	}

	private void _readFromSocket(DatagramPacket packet) {
		byte[] buffer = packet.getData();
		int amt = packet.getLength();
		assert (Thread.currentThread() == _processThread);
		if (amt > 0) {
			if (buffer[0] == REMOTE_MSG_ID_RESPONSE) {
				if (debug) Log.v(TAG, "Got ID response");
				if (amt == 3) {
					if (_connected) {
						if (debug) Log.v(TAG, "Already connected; ignoring ID response");
						return;
					}
					try {
						_socket.connect(packet.getSocketAddress());
					} catch (SocketException e) {
						LogThread.log("Error on socket connect", e);
						Log.e(TAG, "Error on socket connect: " + e.getMessage());
					}
					// hooray we have an id.. we're now officially connected
					_id = buffer[1];
					_nextState = 0; // start over with this ID
					_connected = true;
					_shouldPrintConnected = true;
					assert (_id != -1);
				} else
					Log.e(TAG, "INVALID ID RESPONSE!");
			} else if (buffer[0] == REMOTE_MSG_STATE_ACK) {
				if (amt == 2) {
					long time = SystemClock.uptimeMillis();
					// take note of the next state they want...
					// move ours up to that point (if we havn't yet)
					if (debug) Log.v(TAG, "GOT STATE ACK TO " + (int) (buffer[1] & 0xFF));
					int stateDiff = (buffer[1] - _requestedState) & 0xFF; // upcast
																			// to
																			// make
																			// positive
					assert (stateDiff >= 0);
					if (stateDiff > 0 && stateDiff < 128) {
						_requestedState = (_requestedState + (stateDiff - 1)) % 256;
						long lag = time - _stateBirthTimes[_requestedState];
						if (lag > _currentLag)
							_currentLag = lag;
						_requestedState = (_requestedState + 1) % 256;
						assert (_requestedState == (buffer[1] & 0xFF));
					}
				}
			} else if (buffer[0] == REMOTE_MSG_DISCONNECT_ACK) {
				if (debug) Log.v(TAG, "GOT DISCONNECT ACK!");
				if (amt == 1) {
					_connected = false;
					finish();
				}
			} else if (buffer[0] == REMOTE_MSG_DISCONNECT) {
				if (amt == 2) {
					// ignore disconnect msgs for the first second or two in case we're
					// doing a quick disconnect/reconnect and some old ones come trickling in
					{
						String msg;
						if (buffer[1] == BS_REMOTE_ERROR_VERSION_MISMATCH) {
							msg = "Version mismatch.\nMake sure BombSquad and BombSquad Remote\nare the latest versions and try again.";
						} else if (buffer[1] == BS_REMOTE_ERROR_GAME_SHUTTING_DOWN) {
							msg = "The game has shut down.";
						} else if (buffer[1] == BS_REMOTE_ERROR_NOT_ACCEPTING_CONNECTIONS) {
							msg = "The game is full or not accepting connections.";
						} else {
							msg = "Disconnected by server.";
						}
						class ToastRunnable implements Runnable {
							String s;
							ToastRunnable(String sIn) {
								s = sIn;
							}
							public void run() {
								Context context = getApplicationContext();
								CharSequence text = s;
								int duration = Toast.LENGTH_LONG;
								Toast toast = Toast.makeText(context, text,
										duration);
								toast.show();
							}
						}
						runOnUiThread(new ToastRunnable(msg));
						_connected = false;
						finish();
					}
				}
			}
		}
	}

}
